/*
 * Copyright:      Fabian Maximilian Thiele  mailto:me@apetech.de
 * Author:         Fabian Maximilian Thiele
 * Remarks:        this Copyright must be included
 * known Problems: none
 * Version:        1.0
 * Description:    Real-Time-Clock mit Wochentagsberechnung.
 * 
 */

#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <inttypes.h>
#include <stdlib.h>

#include "rtc.h"
#ifdef USE_SCHEDULER
#include "scheduler.h"
#endif

// RTC Time
volatile date_t rtcTime __attribute__((section(".noinit")));
uint8_t rtcTimeInit __attribute__((section(".noinit")));
volatile int16_t rtcCorrectFact = 0;
volatile uint8_t rtcTimeCorrected = 0;


/*
 * RTC Timer Interrupt.
 * Hier wird die Uhr hochgezhlt.
 * Wenn der Scheduler verwendet wird, wird hier auch die Scheduler-Zeit
 * inkrementiert. Getrennte Zeiten fr RTC und Scheduler sind notwendig, damit
 * zur Laufzeit die Uhr gestellt werden kann, ohne das der Scheduler verrckt
 * spielt.
 */
SIGNAL(SIG_OUTPUT_COMPARE1A) {
	// Timer zurcksetzen
	TCNT1 -= RTC_TCOMP_VAL;
	
	// Zeit inkrementieren
	if(++rtcTime.subSec >= RTC_IRQS_SEC) {
		rtcTime.subSec = 0;
		rtcTime.sec++;
	}
	if(rtcTime.sec >= 60) {
		rtcTime.sec = 0;
		rtcTime.min++;
	}
	if(rtcTime.min >= 60) {
		rtcTime.min = 0;
		rtcTime.hour++;
	}
	if(rtcTime.hour >= 24) {
		rtcTime.hour = 0;
		rtcTime.day++;
	}
	if(rtcTime.day >= rtcDaysOfMonth(rtcTime.month, rtcTime.year)) {
		rtcTime.day = 0;
		rtcTime.month++;
	}
	if(rtcTime.month >= 12) {
		rtcTime.month = 0;
		rtcTime.year++;
	}
	
	// Uhrzeit Korrektur alle 3 Stunden
	// alles andere als schn...
	// wird besser wenn eine 64-Bit Int verwendet wird
	if(!rtcTimeCorrected &&
				rtcTime.sec == 30 &&
				rtcTime.min == 0 &&
				rtcTime.hour%3 == 0) {
	
		if(rtcCorrectFact >= 0) {
			rtcTime.subSec = (rtcCorrectFact/(1000/RTC_IRQS_SEC))%RTC_IRQS_SEC;
			rtcTime.sec += (rtcCorrectFact/(1000/RTC_IRQS_SEC))/RTC_IRQS_SEC;
		} else {
			rtcTime.subSec = RTC_IRQS_SEC+(rtcCorrectFact/(1000/RTC_IRQS_SEC))%RTC_IRQS_SEC;
			rtcTime.sec += (rtcCorrectFact/(1000/RTC_IRQS_SEC))/RTC_IRQS_SEC-1;
		}
		rtcTimeCorrected = 1;
	}
	if(rtcTime.min == 1) {
		rtcTimeCorrected = 0;
	}
	
	#ifdef USE_SCHEDULER
		// Scheduler-Zeit inkrementieren
		if(++schedulerTime.subSec >= RTC_IRQS_SEC) {
			schedulerTime.subSec = 0;
			schedulerTime.sec++;
		}
		if(schedulerTime.sec >= 60) {
			schedulerTime.sec = 0;
			schedulerTime.min++;
		}
		if(schedulerTime.min >= 60) {
			schedulerTime.min = 0;
			schedulerTime.hour++;
		}
		if(schedulerTime.hour >= 24) {
			schedulerTime.hour = 0;
			schedulerTime.day++;
		}
		if(schedulerTime.day >= rtcDaysOfMonth(schedulerTime.month, schedulerTime.year)) {
			schedulerTime.day = 0;
			schedulerTime.month++;
		}
		if(schedulerTime.month >= 12) {
			schedulerTime.month = 0;
			schedulerTime.year++;
		}
	#endif
}

/**
 * Initialisiert die Real Time Clock.
 */
void rtcInit(void) {
	// Datum schon initialisert (z.B. nach einem WDT Reset)?
	if(rtcTimeInit != 0xAA) {
		rtcTime.subSec = 0;
		rtcTime.sec = 0;
		rtcTime.min = 0;
		rtcTime.hour = 0;
		rtcTime.day = 0;
		rtcTime.month = 0;
		rtcTime.year = 2005;
		rtcTimeInit = 0xAA;
	}
	rtcCorrectFact = 0;
	
	// Timer initialisieren
	OCR1A = RTC_TCOMP_VAL;
	TIMSK |= (1<<OCIE1A);
	TCCR1B = RTC_TPS_VAL;
	sei();
}

/**
 * Vergleicht einen beliebigen Zeitpunkt mit der RTC-Zeit.
 * Ist der bergebene Zeitpunkt frher als die momentane RTC-Zeit wird -1
 * zurckgeben. Ist der Zeitpunkt spter als die momentane Zeit wird 1
 * zurckgeben. Sind beide Zeiten gleich ist der Rckgabewert 0.
 */
int8_t rtcCompToSystime(date_t* date) {
	if(rtcTime.year > (*date).year) return -1;
	else if(rtcTime.year < (*date).year) return 1;
	
	else if(rtcTime.month > (*date).month) return -1;
	else if(rtcTime.month < (*date).month) return 1;
	
	else if(rtcTime.day > (*date).day) return -1;
	else if(rtcTime.day < (*date).day) return 1;
	
	else if(rtcTime.hour > (*date).hour) return -1;
	else if(rtcTime.hour < (*date).hour) return 1;
	
	else if(rtcTime.min > (*date).min) return -1;
	else if(rtcTime.min < (*date).min) return 1;
	
	else if(rtcTime.sec > (*date).sec) return -1;
	else if(rtcTime.sec < (*date).sec) return 1;
	
	else if(rtcTime.subSec > (*date).subSec) return -1;
	else if(rtcTime.subSec < (*date).subSec) return 1;
	
	return 0;
}

/**
 * Fgt einem Datum ein bestimmtes Zeitinterval hinzu.
 * incType kann von RTC_SUB_SEC bis RTC_YEAR gehen. value gibt die Anzahl
 * der Zeiteinheiten an, die hinzugefgt werden sollen.
 * 
 * TODO: Zur Zeit knnen nur Zeitintervalle < Tage abgezogen werden
 */
void rtcDateAdd(date_t* date, uint8_t incType, int32_t value) {

	if(incType == RTC_SUB_SEC) {
		(*date).subSec += value%RTC_IRQS_SEC;
		value /= RTC_IRQS_SEC;
		if((*date).subSec >= RTC_IRQS_SEC) {
			(*date).subSec -= RTC_IRQS_SEC;
			value++;
		} else if((*date).subSec < 0) {
			(*date).subSec += RTC_IRQS_SEC;
			value--;
		}
		if(!value) return;
		incType++;
	}
	if(incType == RTC_SEC) {
		(*date).sec += value%60;
		value /= 60;
		if((*date).sec >= 60) {
			(*date).sec -= 60;
			value++;
		} else if((*date).sec < 0) {
			(*date).sec += 60;
			value--;
		}
		if(!value) return;
		incType++;
	}
	if(incType == RTC_MIN) {
		(*date).min += value%60;
		value /= 60;
		if((*date).min >= 60) {
			(*date).min -= 60;
			value++;
		} else if((*date).min < 0) {
			(*date).min += 60;
			value--;
		}
		if(!value) return;
		incType++;
	}
	if(incType == RTC_HOUR) {
		(*date).hour += value%24;
		value /= 24;
		if((*date).hour >= 24) {
			(*date).hour -= 24;
			value++;
		} else if((*date).hour < 0) {
			(*date).hour += 24;
			value--;
		}
		if(!value) return;
		incType++;
	}
	if(incType == RTC_DAY) {
		// Bin mir nich sicher ob das mit den unterschiedlich langen Monaten so
		// funktioniert...
		uint8_t dom = rtcDaysOfMonth((*date).month, (*date).year);
		int8_t months = 0, year;
		while(value > dom) {
			value -= dom;
			months++;
			year = months/12;
			dom = rtcDaysOfMonth((*date).month, (*date).year);
		}
		(*date).day += value;
		if((*date).day >= dom) {
			(*date).day -= dom;
			months++;
		}
		value = months;
		if(!value) return;
		incType++;
	}
	if(incType == RTC_MONTH) {
		(*date).month += value % 12;
		value /= 12;
		
		if((*date).month >= 12) {
			(*date).month -= 12;
			value++;
		}
		uint8_t dom = rtcDaysOfMonth((*date).month, (*date).year+value);
		if((*date).day >= dom) {
			(*date).day -= dom;
			(*date).month++;
		}
		if((*date).month >= 12) {
			(*date).month -= 12;
			value++;
		}
		if(!value) return;
		incType++;
	}
	if(incType == RTC_YEAR) {
		(*date).year += value;
	}
}

/**
 * Gibt den Wochentag eines beliebigen Datums zurck. Die Berechnung erfolgt
 * nach dem Doomsday-Algorithmus.
 * Die zurckgebenen Werte gehen von RTC_MONDAY bis RTC_SUNDAY.
 */
uint8_t rtcDayOfWeek(date_t* date) {
	uint8_t ddayCent=0, m=(*date).month, y=(*date).year;
	int8_t d=(*date).day+2;		// warum auch immer +2...
	
	// Doomsday des ersten Jahres des Jahrhunderts ermitteln
	switch((y%400)/100) {
		case 0: ddayCent = RTC_TUESDAY; break;
		case 1: ddayCent = RTC_SUNDAY; break;
		case 2: ddayCent = RTC_FRIDAY; break;
		case 3: ddayCent = RTC_WEDNESDAY; break;
	}
	
	// Doomsday des Jahres ermitteln
	uint8_t dday = 0, yC = y%100;
	dday = yC%12;
	dday += dday / 4;
	dday += yC/12;
	dday %= 7;
	dday = (ddayCent+dday)%7;
	
	// Doomsday des Monats ermitteln
	uint8_t ddayM = 0;
	switch(m) {
		case  0:
			if(rtcLeapYear(y)) ddayM = 4;
			else ddayM = 3;
			break;
		case  1:
			if(rtcLeapYear(y)) ddayM = 29;
			else ddayM = 28;
			break;
		case  2: ddayM = 7; break;
		case  3: ddayM = 4; break;
		case  4: ddayM = 2; break;
		case  5: ddayM = 6; break;
		case  6: ddayM = 4; break;
		case  7: ddayM = 1; break;
		case  8: ddayM = 5; break;
		case  9: ddayM = 3; break;
		case 10: ddayM = 7; break;
		case 11: ddayM = 5; break;
	}
	d = (d-ddayM);
	if(d<0) d = (d%7)+7;
	d = (d+dday)%7;
	
	return d;
}

/**
 * Gibt die Anzahl der Tage eines Monats zurck.
 * Bei der Anzahl der Tage im Februar werden Schaltjahre bercksichtigt.
 */
uint8_t rtcDaysOfMonth(uint8_t month, uint16_t year) {
	switch(month) {
		case  0: return 31;
		case  1:
			// Schaltjahr?
			if(rtcLeapYear(year))
				 return 29;
			else
				 return 28;
		case  2: return 31;
		case  3: return 30;
		case  4: return 31;
		case  5: return 30;
		case  6: return 31;
		case  7: return 31;
		case  8: return 30;
		case  9: return 31;
		case 10: return 30;
		default: return 31;
	}
}

/**
 * Ermittelt ob das bergebene Jahr ein Schaltjahr ist.
 */
uint8_t rtcLeapYear(uint16_t y) {
	uint8_t b = 0;
	if(y % 4 == 0) b = 1;
	if(y % 100 == 0) b = 0;
	if(y % 400 == 0) b = 1;
	return b;
}

/**
 * Wartet ms Millisekunden. Die Genauigkeit hngt dabei von der kleinsten
 * Zeiteinheit der RTC ab. Ist diese 1/100 s werden mindestens 10ms gewartet.
 */
void rtcWaitMS(uint16_t ms) {
	date_t t = rtcTime;
	rtcDateAdd(&t, RTC_SUB_SEC, ms/(1000/RTC_IRQS_SEC));
	while(rtcCompToSystime(&t) == 1) {
		wdt_reset();
	}
}