#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <stdio.h>
#include <inttypes.h>

#include "uart0.h"

static char uart0RxBuffer[UART0_RX_BUFFER_SIZE];
volatile uint8_t uart0RxRdIndex = 0;
volatile uint8_t uart0RxWrIndex = 0;
volatile uint8_t uart0RxDataLength = 0;

static char uart0TxBuffer[UART0_TX_BUFFER_SIZE];
volatile uint8_t uart0TxRdIndex = 0;
volatile uint8_t uart0TxWrIndex = 0;
volatile uint8_t uart0TxDataLength = 0;

FILE* uart0IO;
volatile uint8_t uart0Flags;

uint16_t uart0Timeout;

/**
 * UART Empfangs-Interrupt
 */
SIGNAL(UART0_RECV_IRQ) {
	char c = UART0_UDR;
	
	if(uart0Flags & (1 << UART0_ECHO) && uart0TxDataLength < UART0_TX_BUFFER_SIZE) {
		uart0Put('\r');
		uart0Put(c);
		uart0Put(' ');
	}
	
	// Empfangsbuffer voll?
	if(uart0RxDataLength >= UART0_RX_BUFFER_SIZE-1) {
		// Overflow-Flag setzen
		uart0Flags |= 1 << UART0_BUF_OV;
	
	// bertragung zu Ende?
	} else if (c != '\r') {
		// Sonst Zeichen speichern
		uart0RxBuffer[uart0RxWrIndex++] = c;
		// ggf. Schreib-Index umbrechen
		if(uart0RxWrIndex >= UART0_RX_BUFFER_SIZE) {
			uart0RxWrIndex = 0;
		}
		// Daten-Counter inkrementieren
		uart0RxDataLength++;
	}
	
	// Nachricht zu Ende?
	if(c == '\r') {
		// String-Terminator, ans Ende.
		uart0RxBuffer[uart0RxWrIndex] = 0;
		// Receive Flag setzen
		uart0Flags |= 1 << UART0_RECV;
	}
}

/**
 * (Ext-)UART Sende-Interrupt
 */
SIGNAL(UART0_TRANS_IRQ) {
	// Weitere zu Sendende Zeichen im Buffer?
	if(uart0TxDataLength) {
		// nchstes Zeichen senden
		UART0_UDR = uart0TxBuffer[uart0TxRdIndex++];
		// ggf. Lese-Index umbrechen
		if(uart0TxRdIndex >= UART0_TX_BUFFER_SIZE) {
			uart0TxRdIndex = 0;
		}
		// Daten-Counter dekrementieren
		uart0TxDataLength--;
	} else {
		// UDR Interrupt deaktivieren
		UART0_UCSRB &= ~(1 << UDRIE);
	}
}

/**
 * Gibt ein Zeichen ber das externe UART aus. Das Zeichen wird entweder
 * sofort gesendet, oder wird im Buffer gespeichert wenn sich bereits zu
 * Sendende Zeichen im Buffer befinden.
 */
int uart0Put(char c) {
	// Bei vollem Buffer warten bis wieder Platz ist.
	if(uart0TxDataLength >= UART0_TX_BUFFER_SIZE) {
		/*date_t toTime = rtcTime;
		if(uart0Timeout) {
			rtcDateAdd(&toTime, RTC_SUB_SEC, uart0Timeout);
		}*/
		
		// warten, solange Buffer voll
		while(uart0TxDataLength >= UART0_TX_BUFFER_SIZE) {
			// Watchdog Timer zurcksetzen
			wdt_reset();
			
			// Bei Timeout Misserfolg zurckgeben
			/*if(uart0Timeout && rtcCompToSystime(&toTime) < 0) {
				return -1;
			}*/
		}
	}
	
	// Interrupts deaktivieren
	cli();
	// Hardware- und Software-Buffer leer?
	if((UART0_UCSRA & (1 << UDRE)) && (uart0TxDataLength == 0)) {
		// Schreibvorgang initiieren
		UART0_UDR = c;
	} else {
		// Sonst Zeichen im Buffer speichern
		uart0TxBuffer[uart0TxWrIndex++] = c;
		// ggf. Schreib-Index umbrechen
		if(uart0TxWrIndex >= UART0_TX_BUFFER_SIZE) {
			uart0TxWrIndex = 0;
		}
		// Daten-Counter inkrementieren
		uart0TxDataLength++;
	}
	// Date Register Empty Interrupt aktivieren
	UART0_UCSRB |= (1 << UDRIE);
	// Interrupts wieder aktivieren
	sei();
	
	// Erfolg zurckgeben
	return 0;
}

/**
 * Die Funktion gibt ein empfangenes Zeichen zurck. Falls sich Zeichen im
 * Empfangsbuffer befinden werden zunchst diese zurckgegeben. Ist der Buffer
 * leer wird auf ein Zeichen gewartet.
 */
int uart0Get(void) {
	int c;
	
	// Bei leerem Buffer auf Zeichen warten
	if(!uart0RxDataLength) {
		/*date_t toTime = rtcTime;
		if(uart0Timeout) {
			rtcDateAdd(&toTime, RTC_SUB_SEC, uart0Timeout);
		}*/
		
		// Auf Zeichen warten
		while(!uart0RxDataLength) {
			// Watchdog Timer zurcksetzen
			wdt_reset();
			
			// Bei Timeout Misserfolg zurckgeben
			/*if(uart0Timeout && rtcCompToSystime(&toTime) < 0) {
				return -1;
			}*/
		}
	}
	
	// Interrupts deaktivieren
	cli();
	// Zeichen vom Buffer holen
	c = uart0RxBuffer[uart0RxRdIndex++];
	// ggf. Schreib-Index umbrechen
	if(uart0RxRdIndex >= UART0_RX_BUFFER_SIZE) {
		uart0RxRdIndex = 0;
	}
	// Zhler dekrementieren
	uart0RxDataLength--;
	// Interrupts wieder aktivieren
	sei();
	
	return c;
}

/**
 * Verwirdt den Inhalt des Empfangsbuffers.
 */
void uart0FlushRxBuffer(void) {
	cli();
	uart0RxDataLength = 0;
	uart0RxRdIndex = uart0RxWrIndex;
	sei();
}

/**
 * Verwirdt den Inhalt des Sendebuffers.
 */
void uart0FlushTxBuffer(void) {
	cli();
	uart0TxDataLength = 0;
	uart0TxRdIndex = uart0TxWrIndex;
	sei();
}

/**
 * Initialisiert das Externe UART (UART0).
 * Es werden die RX Complete und Data Register Empty Interrupts verwendet.
 * Baudrate, Empfangs- und Sendebuffer Gre werden im Headerfile
 * definiert.
 */
void uart0Init(void) {
	// Hardware initialisieren
	UART0_UCSRA = 0;
	UART0_UCSRB = (1 << RXCIE) |		// Receiver IRQ Enable
				(1 << RXEN) |			// Receiver Enable
				(1 << TXEN);			// Transmitter Enable
	UART0_UCSRC = 0x06;					// 8 Datenbits, 1 Stoppbit, No parity
	UART0_UBRR = F_OSC/(16*UART0_BAUDRATE)-1;
	
	uart0Flags |= (1 << UART0_ECHO);	// Echo enabled
	uart0Timeout = 0;					// Timeout deaktiviert
	
	sei();								// Global Interrupt Enable
	
	// IO-Stream fr Nutzung mit stdio ffnen
	uart0IO = fdevopen(uart0Put, uart0Get, 0);
}

/**
 * Deaktiviert das UART.
 */
void uart0Disable(void) {
	// Hardware abschalten
	UART0_UCSRB = 0;
}
